// Cory McWilliams
// mcwilc@rpi.edu
// 10/15/2004

#include <stdlib.h>

// This should be (but doesn't seem to be in VCC South) defined by math.h/cmath
#ifndef M_PI
#define M_PI 3.1415926535
#endif

#include <GL/glut.h>
#include <cmath>
#include <iostream>
#include <string>
#include <time.h>
#include "Humanoid.h"
#include "proj2_support.h"
#include "GroundTexture.h"
#include "NotKitten.h"

#define BOX_COUNT 20

int gWidth;
int gHeight;
bool gKeys[256];
bool gButtons[3];
bool gArrows[4];
const int gDelay = 15;
float gPos[3];
float gDistance;
float gCharacterRotY = 0.0f;
float gCameraRotY = 0.0f;
float gCameraRotX = 0.0f;
float gVelocityY;
int gCursor[2];
GLUquadric *gQuadric;
Humanoid gHumanoid;
bool gAnimate;
enum { PLAYING_NONE, PLAYING_WALK, PLAYING_PICK_UP, PLAYING_PICKED_UP } gPlaying = PLAYING_NONE;
unsigned int gTexFloor;
int gHoldingIndex = -1;
struct {
	float x;
	float y;
	float r;
	float c[4];
} gBox[BOX_COUNT];
std::string gText;
int gTextIndex;
bool gDone = false;

void box();

void drawString(const char *string) {
	const char *c = string;
	while (*c) {
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *c++);
	}
}

void reset() {
	gText = "You are robot.  Your job is to find kitten.";
	gDone = false;
	gArrows[0] = gArrows[1] = gArrows[2] = gArrows[3] = false;
	gPos[0] = 50.0f;
	gPos[1] = 0.0f;
	gPos[2] = 50.0f;
	gCharacterRotY = 0.0f;
	gCameraRotY = 0.0f;
	gCameraRotX = 0.0f;
	gAnimate = false;
	gPlaying = PLAYING_NONE;
	gDistance = 16.0f;
	gHumanoid.reset();
	gHoldingIndex = -1;
	gHumanoid.hold(0);

	for (int i = 0; i < BOX_COUNT; i++) {
		gBox[i].x = rand() / (RAND_MAX + 1.0f) * 100.0f;
		gBox[i].y = rand() / (RAND_MAX + 1.0f) * 100.0f;
		gBox[i].r = rand() / (RAND_MAX + 1.0f) * 90.0f;
		gBox[i].c[0] = rand() / (RAND_MAX + 1.0f);
		gBox[i].c[1] = rand() / (RAND_MAX + 1.0f);
		gBox[i].c[2] = rand() / (RAND_MAX + 1.0f);
		gBox[i].c[3] = 1.0f;
	}

	gTextIndex = static_cast<int>(rand() / (RAND_MAX + 1.0f) * gNotKittenCount);
}


void init() {
	srand(time(0));

	std::cout <<
	"In this game, you are robot.  Your job is to find kitten.  This task is\n"
	"complicated by the existence of various things which are not kitten.  Robot\n"
	"must stand over items and press enter to pick them up to determine if they are\n"
	"kitten or not.  The game ends when robotfindskitten.  Alternatively, you may\n"
	"end the game by hitting the j key.\n\n";
	
	glClearColor(0.0f, 0.0f, 0.7f, 0.0f);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_TEXTURE_2D);
	gQuadric = gluNewQuadric();
	glGenTextures(1, &gTexFloor);
	glutIgnoreKeyRepeat(true);
	if (gTexFloor) {
		GLenum format = ((GroundTexture.bytes_per_pixel == 3)?GL_RGB:GL_RGBA);
		glBindTexture(GL_TEXTURE_2D, gTexFloor);
		gluBuild2DMipmaps(GL_TEXTURE_2D, format, GroundTexture.width, GroundTexture.height, format, GL_UNSIGNED_BYTE, GroundTexture.pixel_data);
	}
	reset();
}

void displayText() {
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glColor3f(1.0f, 1.0f, 1.0f);
	glRasterPos2f(0.01f, 0.01f);
	glutReportErrors();
	drawString(gText.c_str());
	glutReportErrors();
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void display() {
	const float blockSize = 10.0f;
	float mostlyWhite[] = {0.8f, 0.8f, 0.8f, 1.0f};
	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mostlyWhite);

	glPushMatrix();
		glTranslatef(0.0f, 0.0f, -gDistance);
		glRotatef(-gCameraRotX, 1.0f, 0.0f, 0.0f);
		glRotatef(-gCameraRotY, 0.0f, 1.0f, 0.0f);

		glPushMatrix();
		glRotatef(gCharacterRotY, 0.0f, 1.0f, 0.0f);
		gHumanoid.draw();
		glPopMatrix();

		glTranslatef(-gPos[0], -gPos[1], -gPos[2]);

		glBindTexture(GL_TEXTURE_2D, gTexFloor);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glPushMatrix();
		glTranslatef(0.0f, -(HIP_HEIGHT/2+UPPER_LEG_HEIGHT+LOWER_LEG_HEIGHT), 0.0f);
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mostlyWhite);
		glBegin(GL_QUADS);
		for (int i = 0; i < 10; i++) {
			for (int j = 0; j < 10; j++) {
				glTexCoord2f(0.0f, 0.0f);
				glVertex3f(blockSize*i+0.0f, 0.0f, blockSize*j+0.0f);
				glTexCoord2f(0.0f, 1.0f);
				glVertex3f(blockSize*i+0.0f, 0.0f, blockSize*j+blockSize);
				glTexCoord2f(1.0f, 1.0f);
				glVertex3f(blockSize*i+blockSize, 0.0f, blockSize*j+blockSize);
				glTexCoord2f(1.0f, 0.0f);
				glVertex3f(blockSize*i+blockSize, 0.0f, blockSize*j+0.0f);
			}
		}
		glEnd();

		for (int i = 0; i < BOX_COUNT; i++) {
			if (i != gHoldingIndex) {
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gBox[i].c);
				glPushMatrix();
				glTranslatef(gBox[i].x, 0.5f, gBox[i].y);
				glRotatef(gBox[i].r, 0.0f, 1.0f, 0.0f);
				glutSolidCube(1.0f);
				glPopMatrix();
			}
		}
		
		glPopMatrix();
		glBindTexture(GL_TEXTURE_2D, 0);

		float lightPosition[] = {100.0f, 100.0f, 100.0f, 1.0f};
		glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);

	glPopMatrix();
	
	displayText();
	
	glutSwapBuffers();
}


void reshape(int width, int height) {
	glViewport(0, 0, gWidth = width, gHeight = height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, float(width)/float(height?height:1.0f), 0.5f, 300.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void quit() {
	gluDeleteQuadric(gQuadric);
}

void translateCamera(float dx, float dy, float dz) {
	double theta = (gCharacterRotY) * M_PI / 180.0;
	float sint = sin(theta);
	float cost = cos(theta);
	gPos[0] += dz * sint + dx * cost;
	gPos[1] += dy;
	gPos[2] += dz * cost - dx * sint;
}

void movementUpdate() {
	const float velocity = 0.25f;
	const float rotspeed = 3.0f;

	const char partKey[HUMANOID_JOINT_COUNT] = {
		'r','e','d','q','a','z','y','g','v','s','x','f','c'
	};

	for (int i = 0; i < HUMANOID_JOINT_COUNT; i++) {
		if (gKeys[partKey[i]]) {
			gHumanoid.joint(i, gHumanoid.joint(i)+1.0f);
		} else if (gKeys[toupper(partKey[i])]) {
			gHumanoid.joint(i, gHumanoid.joint(i)-1.0f);
		}
	}

	gVelocityY -= 0.005f;
	gPos[1] += gVelocityY;
	if (gPos[1] < 0.0f) {
		gPos[1] = 0.0f;
		gVelocityY = 0.0f;
	}

	if (gArrows[0]) translateCamera(0.0f, 0.0f, velocity); // up
	if (gArrows[2]) translateCamera(0.0f, 0.0f, -velocity); // down

	if (gArrows[1]) gCharacterRotY -= 5.0f;
	if (gArrows[3]) gCharacterRotY += 5.0f;

	if (gPlaying == PLAYING_WALK && !gArrows[0] && !gArrows[1] && !gArrows[2] && !gArrows[3]) {
		gAnimate = false;
		gPlaying = PLAYING_NONE;
		gHumanoid.reset(); // Play some idle animation.
	}

	if (gPos[0] > 100.0f) gPos[0] = 100.0f;
	if (gPos[0] < 0.0f) gPos[0] = 0.0f;
	if (gPos[2] > 100.0f) gPos[2] = 100.0f;
	if (gPos[2] < 0.0f) gPos[2] = 0.0f;
}

void keyboard(unsigned char key, int x, int y) {
	if (gDone) {
		reset();
	}
	gKeys[key] = true;

	switch (key) {
		case ' ':
			if (gPos[1] <= 0.0f) {
				gVelocityY = 0.1f;
			}
			break;
		case 'i':
		case 'I':
			reset();
			break;
		case 'l': {
			std::string filename;
			std::cout << "Load animation: ";
			std::getline(std::cin, filename);
			gHumanoid.loopAnimation(filename.c_str());
			gAnimate = true;
			}
			break;
		case '\r':
			gHumanoid.loopAnimation("pickup.txt");
			gAnimate = true;
			gPlaying = PLAYING_PICK_UP;
			break;
		case 'p':
			gAnimate = !gAnimate;
			break;
		case 'o':
			gHumanoid.dump();
			break;
		case 'j':
			exit(0);
			break;
	}

	movementUpdate();
}

void keyboardUp(unsigned char key, int x, int y) {
	if (gDone) {
		reset();
	}
	gKeys[toupper(key)] = false;
	gKeys[tolower(key)] = false;
}

void special(int key, int x, int y) {
	if (gDone) {
		reset();
	}
	if (key == GLUT_KEY_UP) {
		gArrows[2] = false;
		gArrows[0] = true;
	}
	if (key == GLUT_KEY_RIGHT) {
		gArrows[1] = true;
	}
	if (key == GLUT_KEY_DOWN) {
		gArrows[0] = false;
		gArrows[2] = true;
	}
	if (key == GLUT_KEY_LEFT) {
		gArrows[3] = true;
	}
	if (gPlaying != PLAYING_WALK && (key == GLUT_KEY_UP || key == GLUT_KEY_DOWN)) {
		gHumanoid.loopAnimation("walk.txt");
		gAnimate = true;
		gPlaying = PLAYING_WALK;
	}
}

void specialUp(int key, int x, int y) {
	if (gDone) {
		reset();
	}
	if (key == GLUT_KEY_UP) gArrows[0] = false;
	if (key == GLUT_KEY_RIGHT) gArrows[1] = false;
	if (key == GLUT_KEY_DOWN) gArrows[2] = false;
	if (key == GLUT_KEY_LEFT) gArrows[3] = false;
}

void timer(int t) {
	glutTimerFunc(gDelay, timer, 0);

	if (gAnimate) {
		gHumanoid.step();

		if (gPlaying == PLAYING_PICK_UP && gHumanoid.lastFrame()) {
			gHumanoid.loopAnimation("pickedup.txt");
			gAnimate = true;
			gPlaying = PLAYING_PICKED_UP;

			int nearBox = -1;

			for (int i = 0; i < BOX_COUNT; i++) {
				if (((gPos[0] - gBox[i].x) * (gPos[0] - gBox[i].x) + (gPos[2] - gBox[i].y) * (gPos[2] - gBox[i].y)) < 4.0f) {
					nearBox = i;
				}
			}
			
			if (gHoldingIndex == -1 && nearBox != -1) {
				gHoldingIndex = nearBox;
				gHumanoid.hold(box);
				if (nearBox == 0) {
					gDone = true;
					gText = "Congratulations, robot!  You found kitten!";
					gHumanoid.loopAnimation("walk.txt");
					gAnimate = true;
					gPlaying = PLAYING_WALK;
					gArrows[0] = true;
					gArrows[3] = true;
					gArrows[2] = false;
					gArrows[1] = false;
				} else {
					gText = gNotKitten[(gTextIndex + nearBox) % gNotKittenCount];
				}
			} else if (gHoldingIndex != -1)  {
				gBox[gHoldingIndex].x = gPos[0];
				gBox[gHoldingIndex].y = gPos[2];
				gBox[gHoldingIndex].r = gCharacterRotY;
				gHoldingIndex = -1;
				gHumanoid.hold(0);
				gText = "";
			}
		} else if (gPlaying == PLAYING_PICKED_UP && gHumanoid.lastFrame()) {
			if (gKeys[0] || gKeys[2]) {
				gHumanoid.loopAnimation("walk.txt");
				gAnimate = true;
				gPlaying = PLAYING_WALK;
			} else {
				gHumanoid.reset();
				gAnimate = false;
				gPlaying = PLAYING_NONE;
			}
		}
	}
	movementUpdate();

	glutPostRedisplay();
}

void mouse(int button, int state, int x, int y) {
	gCursor[0] = x;
	gCursor[1] = y;
	if (button >= 0 && button <= 2) {
		gButtons[button] = state == GLUT_DOWN;
	}
}

void motion(int x, int y) {
	if (gButtons[0]) {
		gCameraRotY -= (180.0f * (x - gCursor[0])) / gWidth;
		gCameraRotX -= (180.0f * (y - gCursor[1])) / gHeight;
	} else if (gButtons[2]) {
		gDistance -= (10.0f * (y - gCursor[1])) / gHeight;
	}
	gCursor[0] = x;
	gCursor[1] = y;
}

void box() {
	glPushAttrib(GL_LIGHTING_BIT);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gBox[gHoldingIndex].c);
	glutSolidCube(1.0f);
	glPopAttrib();
}

int main(int argc, char *argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE|GLUT_DEPTH);
	glutInitWindowSize(gWidth = 640, gHeight = 480);
	glutCreateWindow(argv[0]);
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutSpecialFunc(special);
	glutSpecialUpFunc(specialUp);
	glutTimerFunc(gDelay, timer, 0);
	glutReshapeFunc(reshape);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	atexit(quit);
	init();
	glutMainLoop();

	return 0;
}
